unit a3d;

interface
uses FormGL, Windows, Forms, Classes;
type
  PTextInfo = ^TTextInfo; 
  TTextInfo = record
    InterfaceVersion:integer;
    FullDriverName,ShortDriverName,SysDriverID,DriverVersion:ShortString;
  end;

  TFD = record
    Vertex,Normal,MaterialID:integer;
  end;
  TFace = array of TFD;

  TA3DColor = TGLColor;

  TAbstract3DFrame = class
    protected
      FFrames:TList; // child frames
      FParent:TAbstract3DFrame;
      function GetFrCount:integer;
      function GetFr(index:integer):TAbstract3DFrame;virtual;
      procedure Process;virtual;
    public
      property FrameCount:integer read GetFrCount;
      property Frames[Index:integer]:TAbstract3DFrame read GetFR;
//      property Texture:TAbstract3DTexture read Ftex;
      constructor Create(Parent:TAbstract3DFrame);virtual;
      destructor Destroy;override;
      procedure AddFrame(Frame:TAbstract3DFrame);virtual;
      function AddVertex(Vertex:TVector):cardinal;overload;virtual;abstract;
      function AddVertex(x,y,z:Single):cardinal;overload;virtual;abstract;
      function AddNormal(Normal:TVector):cardinal;overload;virtual;abstract;
      function AddNormal(x,y,z:Single):cardinal;overload;virtual;abstract;
      function AddMaterial(Material:TGLMaterial):cardinal;overload;virtual;abstract;
      function AddMaterial(Ambient,Diffuse,Specular:TA3DColor;Emission,Shininess:single):cardinal;overload;virtual;abstract;
      procedure ReplaceMaterial(ID:integer;Material:TGLMaterial);overload;virtual;abstract;
      procedure ReplaceMaterial(ID:integer;Ambient,Diffuse,Specular:TA3DColor;Emission,Shininess:single);overload;virtual;abstract;
      procedure AddFace(FVN:array of TFVN);virtual;abstract;
      procedure SetRotation(x,y,z,increment:Single);virtual;abstract;
      procedure Rotate(x,y,z:Single);virtual;abstract;
//      procedure SetTexture(Texture:TGLTexture;DestroyOld:Boolean = true);virtual;abstract;
      procedure SetPosition(x,y,z:Single);virtual;abstract;
      procedure Move(dx,dy,dz:Single);virtual;abstract;
      function GetPolygonCount:cardinal;virtual;abstract;
  end;

  TAbstract3D = class
    protected
      FParent:TForm;
      FCanDraw:boolean;
      Frenderer:shortstring;
      FColorBits:byte;
      FWidth,FHeight:cardinal;
      FCamX,FCamY,FCamZ,FTgX,FTgY,FTgZ,FUpX,FUpY,FUpZ:Single;
      FUgol:Single;
      Frect:Trect;
      FBKColor:TGLColor;
      FScene:TAbstract3DFrame;
      procedure SetColorBits(value:byte);virtual;
      procedure SetWidth(value:cardinal);virtual;
      procedure SetHeight(value:cardinal);virtual;
    public
      constructor Create(Form:TForm;Rect:TRect);virtual;
//      destructor Destroy;virtual;abstract;
      property ViewRect:TRect read FRect write FRect;
      property CanDraw:boolean read FCanDraw;
      property Renderer:shortstring read Frenderer;
      property BackColor:TGLColor read FBKColor write FBKColor;
      property Ugol:single read FUgol write FUgol;
      property Scene:TAbstract3DFrame read FScene;
      property ColorBits:byte read FColorBits write SetColorBits;
      property Width:cardinal read FWidth write SetWidth;
      property Height:cardinal read FHeight write SetHeight;
      procedure Initialize;virtual;abstract;
      procedure Finalize;virtual;abstract;
      procedure SetCameraPos(CamX,CamY, CamZ,TargetX,TargetY,TargetZ,UpX,UpY,UpZ:Single);virtual;
      procedure Render;virtual;
      procedure ShowSetupWindow;virtual;abstract;
  end;

implementation

constructor TAbstract3D.Create(Form:TForm;Rect:TRect);
begin
  inherited Create;
  FBKColor.r:=0;FBKColor.g:=0;FBKColor.b:=0;FBKColor.a:=0;
end;

procedure TAbstract3D.SetCameraPos(CamX,CamY, CamZ,TargetX,TargetY,TargetZ,UpX,UpY,UpZ:Single);
begin
  FCamX:=CamX; FCamY:=CamY; FCamZ:=CamZ;
  FTGX:=TargetX; FTgY:=TargetY; FTgZ:=TargetZ;
  FUpX:=UpX; FUpY:=UpY; FUpZ:=UpZ;
end;

procedure TAbstract3D.SetColorBits(value:byte);
begin
  FColorBits:=value;
end;

procedure TAbstract3D.SetWidth(value:cardinal);
begin
  FWidth:=value;
end;

procedure TAbstract3D.SetHeight(value:cardinal);
begin
  FHeight:=value;
end;

constructor TAbstract3DFrame.Create(Parent:TAbstract3DFrame);
begin
  inherited Create;
  FParent:=Parent;
  if Assigned(parent) then Parent.FFrames.Add(self);
  FFrames:=TList.Create;
end;

destructor TAbstract3DFrame.Destroy;
var i:integer;
begin
  for i:=0 to FFrames.Count-1 do
    TAbstract3DFrame(FFrames.Items[i]).Free;
  FFrames.Free;
  inherited destroy;
end;

function TAbstract3DFrame.GetFrCount:integer;
begin
  Result:=FFrames.Count;
end;

function TAbstract3DFrame.GetFr(index:integer):TAbstract3DFrame;
begin
  Result:=FFrames.Items[index];
end;

procedure TAbstract3DFrame.Process;
var i:integer;
begin
 for i:=0 to FFrames.Count-1 do
   TAbstract3DFrame(FFrames.Items[i]).Process;
end;

procedure TAbstract3D.Render;
begin
  FScene.Process;
end;

procedure TAbstract3DFrame.AddFrame(Frame:TAbstract3DFrame);
begin 
  if not Assigned(Frame) then Exit;
  if FFrames.IndexOf(Frame)<>-1 then Exit;
  if Assigned(Frame.FParent) then Frame.FParent.FFrames.Remove(Frame);
  FFrames.Add(Frame); 
end;

end.
