{FormGL -  DelphiX(Direct3D abstract mode)}
{ Naumov's corp - by Mish!}
unit FormGL;

interface
 Uses OpenGL, SysUtils, Forms, Classes, Windows, Messages, Graphics, Dialogs;

 const MaxTextureX = 1024;
       MaxTextureY = 1024;

 type

  TGLColor = record
    case integer of
     0:(r,g,b,a:GLFloat);
     1:(d:array[0..3] of GLFloat);
  end;

  TGLForm = class;

  TGLMaterial = record
     Ambient,Diffuse,Specular:TGLColor;
     Emission,Shininess:GLFloat;
  end;

  TVector = record
    x,y,z:Single;
  end;

  TVertex = TVector;
  TNormal = TVector;

  PFVN = ^TFVN;
  TFVN = record
    Vertex,Normal:Integer;
    TexX,TexY:GLFloat;
    Material:cardinal;
  end;

  TGLFaceArr = array of array of TFVN;
  PGLFaceArr = ^TGLFaceArr;
  TGL3VFace = array[1..3] of TFVN;
  TGL4VFace = array[1..4] of TFVN;

  TCoordMode = (cmLocal,cmWorld);

  PGLLight = ^TGLLight;
  TGLLight = record
    Enabled:boolean;
    Position,Direction:TVector;
    Ambient,Diffuse,Specular:TGLColor;
  end;

  TTexData = array[0..MaxTextureX-1,0..MaxTextureY-1,0..3] of GLuByte;
  PTexData = ^TTexData;

  TGLTexture = class
  private
   FrAss:TList;
   FE:boolean;
   FW,FH:integer;
   MemUse:integer;
   Data:PTexData;
   FTr,q:boolean;
   FTrColor:TGLColor;
   procedure SetTransparent(value:Boolean);
   procedure SetTransparentColor(Color:TGLColor);
  public
    property Transparent:boolean read FTr write SetTransparent;
    property TransparentColor:TGLColor read FTrColor write SetTransparentColor;
    property Empty:boolean read Fe default True;
    property Width:integer read FW;
    property Height:integer read FH;
    procedure LoadFromFile(FileName:string);
    procedure Clear;
    constructor Create;
    destructor Destroy;override;
  end;

  TGLHintMode = (glNone,glFast,glBest);

  TGLFrame = class
    private
      FPC,FLS,FPS,FPolS:TGLHintMode;
      FAA:Boolean;
      FMat:TGLMaterial;
      FParent:TGLFrame;
      px,py,pz:Single;// position
      rx,ry,rz,incr,trx,tryy,trz:Single; // rotation
      fsx,fsy,fsz:GLFloat; // scale
      FFrames:TList; // child frames
      FVertex:Array of TVertex;
      FNormals:Array of TNormal;
      FFaces3V:array of array of TFVN;
      FFacesPol:array of array of TFVN;
      FCM:TCoordMode;
      FTex:TGLTexture;
      FNC:boolean;
      FCompiled:boolean;
      FList:integer;
      procedure Process;
      function GetFrCount:integer;
      function GetFr(index:integer):TGLFrame;
      procedure SetNC(value:boolean);
      procedure ProcessPolygons;
    public
      property LineSmooth:TGLHintMode read FLS write FLS default glFast;
      property PointSmooth:TGLHintMode read FPS write FPS default glFast;
      property PolygonSmooth:TGLHintMode read FPolS write FPolS default glFast;
      property PerspectiveCorrection:TGLHintMode read FPC write FPC default glBest;
      property AntiAliasing:Boolean read FAA write FAA default False;
      property NeedCompile:boolean read FNC write SetNC default False;
      property Texture:TGLTexture read Ftex;
      property Material:TGLMaterial read Fmat write Fmat;
      property FrameCount:integer read GetFrCount;
      property Frames[Index:integer]:TGLFrame read GetFR;
      property CoordMode:TCoordMode read FCM write FCM default cmLocal;
      property Compiled:boolean read FCompiled default False;
      procedure Compile;
//      procedure Scale(sx,sy,sz:GLFloat);
      procedure SetRotation(x,y,z,increment:GLFloat);
      procedure Rotate(x,y,z:GLFloat);
      function AddVertex(x,y,z:GLFloat):Integer;
      function AddNormal(x,y,z:GLFloat):Integer;
      function AddFrame(Frame:TGLFrame):integer;
      procedure SetTexture(Texture:TGLTexture;DestroyOld:Boolean = true);
      procedure SetPosition(x,y,z:GLFloat);
      procedure Move(dx,dy,dz:GLFloat);
      procedure AddFace(FVN:array of TFVN);
      constructor Create(Parent:TGLFrame);
      destructor Destroy; override;
  end;

  TGLShadeMode = (glSmooth, glFlat);

  TGLViewPort = class
  private
    FLights:TList;
    FRect:TRect;
    FMainFr:TGLFrame;
    FBKColor:TGLColor;
    FNearZ,FFarZ,FU:GLFloat;
    FCamX,FCamY,FCamZ,FTgX,FTgY,FTgZ,FUpX,FUpY,FUpZ:GLFloat;
    FShMode:TGLShademode;
    procedure Process;
    function GetLgCount:integer;
    function GetLg(index:integer):PGLLight;
  public
    property Ugol:GLFloat read FU write FU;
    property LightCount:integer read GetLgCount;
    property Lights[Index:integer]:PGLLight read GetLg;
    property ViewRect:TRect read Frect write Frect;
    property ShadeMode:TGLShademode read FShMode write FShMode default glSmooth;
    property NearZ:GLFloat read FNearZ write FNearZ;
    property FarZ:GLFloat read FFarZ write FFarZ;
    property BackColor:TGLColor read FBKColor write FBKColor;
    property Scene:TGLFrame read FMainFr;
    function AddLight(Light:TGLLight):integer;overload;
    function AddLight(Light:PGLLight):integer;overload;
    procedure SetCameraPos(CamX,CamY, CamZ,TargetX,TargetY,TargetZ,UpX,UpY,UpZ:GLFloat);
    constructor Create(Form:TGLForm;Rect:TRect);
    destructor Destroy;override;
  end;

  TGLForm = class(TForm)
    private
      DC:HDC;
      hrc:HGLRC;
      Palette:HPalette;
      FColorBits:Byte;
      FDepthBits:Byte;
      FASPF:boolean;
      FViewPorts:TList;
      FCanDraw:boolean;
      function GetVP(index:integer):TGLViewPort;
      procedure SetDCPixelFormat;
      procedure SetColorBits(Value:byte);
      procedure SetDepthBits(Value:byte);
      procedure WMQueryNewPalette(var Msg: TWMQueryNewPalette); message WM_QUERYNEWPALETTE;
      procedure WMPaletteChanged(var Msg: TWMPaletteChanged); message WM_PALETTECHANGED;
      procedure WMPaint(var Msg:TMessage); message WM_PAINT;
    protected
    public
      property CanDraw:boolean read FCanDraw;
      property GLHandle:HGLRC read hrc;
      property ViewPorts[Index:Integer]:TGLViewPort read GetVP;
      function AddViewPort(Rect:TRect):integer;
      procedure DeleteViewPort(ViewPort:TGLViewPort); overload;
      procedure DeleteViewPort(ViewPort:Integer); overload;
      function NewViewPort(Rect:TRect):TGLViewPort;
      function GetViewPortCount:Integer;
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Process;
    published
      property ColorBits:byte read FColorBits write SetColorBits;
      property DepthBits:byte read FDepthBits write SetDepthBits;
      property AutoSetPixelFormat:boolean read FASPF write FASPF default False;
  end;

 const
   GLBlack:TGLColor = (r:0;g:0;b:0;a:1);
   GLGrey:TGLColor = (r:0.5;g:0.5;b:0.5;a:1);
   GLWhite:TGLColor = (r:1;g:1;b:1;a:1);
   DefaultMaterial:TGLMaterial = (Ambient:(r:0.1;g:0.1;b:0.1;a:1);Diffuse:(r:0.5;g:0.5;b:0.5;a:1);Specular:(r:1;g:1;b:1;a:1);Emission:0.0;Shininess:10.0);
   DefaultNearZ:GLFloat = 0.1;
   DefaultFarZ:GLFloat = 100.0;

 function Max(a,b:integer):integer;

implementation

procedure TGLTexture.SetTransparent(value:Boolean);
begin
  FTr:=value;
  if value then begin
     q:=False;
     SetTransparentColor(FTrColor);
  end;
end;

procedure TGLTexture.SetTransparentColor(Color:TGLColor);
var i,j,i1,i2:integer;
begin
  if q then
  if (FTrColor.r=Color.r) and (FTrColor.g=Color.g) and (FTrColor.b=Color.b)then Exit;
  q:=True;
  FTrColor:=Color;
  if not FTr then Exit;
  i1:=0;i2:=0;
  for i:=0 to FH-1 do
   begin
     for j:=0 to FW-1 do
     begin
       if (Data[i1,i2,2]=Color.r) and (Data[i1,i2,1]=Color.g) and (Data[i1,i2,0]=Color.b)
           then Data[i1,i2,3]:=0;
       inc(i2);if i2>(MaxTextureX-1) then begin i2:=0;inc(i1) end;
       if i1>(MaxTextureY-1) then begin i1:=0 end;
     end;
   end;
end;

procedure TGLFrame.SetNC(value:boolean);
begin
  if FNC=value then Exit;
  FNC:=value;
  if not value then begin
     glDeleteLists(Flist,1);
     FCompiled:=False;
  end
  else Compile;
end;

procedure TGLFrame.Compile;
begin
  if FCompiled or (not FNC) then Exit;
  glDeleteLists(FList,1);
  glNewList(FList, GL_Compile);
    ProcessPolygons;
  glEndList;
  FCompiled:=true;
end;

procedure TGLFrame.ProcessPolygons;
var i,j:integer;
begin
  glMaterialf(GL_FRONT,GL_DIFFUSE,FMat.Diffuse.d[0]);
  glMaterialf(GL_FRONT,GL_AMBIENT,FMat.Ambient.d[0]);
  glMaterialf(GL_FRONT,GL_SPECULAR,FMat.Specular.d[0]);
  glMaterialf(GL_FRONT,GL_SHININESS,FMat.Shininess);
  glMaterialf(GL_FRONT,GL_EMISSION,FMat.Emission);
  for i:=0 to High(FFacesPol) do
   begin
     glBegin(GL_POLYGON);
     for j:=0 to high(FFacesPol[i]) do
      begin
       if (j=0) or ((j<>0)and(FFacesPol[i,j-1].Normal<>FFacesPol[i,j].Normal)) then
       glNormal3f(FNormals[FFacesPol[i,j].Normal].x,
                  FNormals[FFacesPol[i,j].Normal].y,
                  FNormals[FFacesPol[i,j].Normal].z);
       if Assigned(FTEx) then glTexCoord2d(FFacesPol[i,j].TexX,FFacesPol[i,j].TexY);
       glVertex3f(FVertex[FFacesPol[i,j].Vertex].x,
                  FVertex[FFacesPol[i,j].Vertex].y,
                  FVertex[FFacesPol[i,j].Vertex].z);
      end;
     glEnd;
   end;
  {processing triangle faces}
  if High(FFaces3V)<>-1 then glBegin(GL_TRIANGLES);
  for i:=0 to High(FFaces3V) do
   begin
     for j:=0 to 2 do
      begin
       if (j=0) or ((j<>0)and(FFaces3V[i,j-1].Normal<>FFaces3V[i,j].Normal)) then
       glNormal3f(FNormals[FFaces3V[i,j].Normal].x,
                  FNormals[FFaces3V[i,j].Normal].y,
                  FNormals[FFaces3V[i,j].Normal].z);
       if Assigned(FTEx) then glTexCoord2d(FFaces3V[i,j].TexX,FFaces3V[i,j].TexY);
       glVertex3f(FVertex[FFaces3V[i,j].Vertex].x,
                  FVertex[FFaces3V[i,j].Vertex].y,
                  FVertex[FFaces3V[i,j].Vertex].z);
      end;
   end;
  if High(FFaces3V)<>-1 then  glEnd;

  {processing quadric faces}
{  if High(FFaces4V)<>-1 then glBegin(GL_QUADS);
  for i:=0 to High(FFaces3V) do
   begin
     for j:=0 to 3 do
      begin
       if (j=0) or ((j<>0)and(FFaces4V[i,j-1].Normal<>FFaces4V[i,j].Normal)) then
       glNormal3f(FNormals[FFaces4V[i,j].Normal].x,
                  FNormals[FFaces4V[i,j].Normal].y,
                  FNormals[FFaces4V[i,j].Normal].z);
       if Assigned(FTEx) then glTexCoord2d(FFaces4V[i,j].TexX,FFaces4V[i,j].TexY);
       glVertex3f(FVertex[FFaces4V[i,j].Vertex].x,
                  FVertex[FFaces4V[i,j].Vertex].y,
                  FVertex[FFaces4V[i,j].Vertex].z);
      end;
   end;
  if High(FFaces4V)<>-1 then  glEnd;}
end;

function DParToGL(par:TGLHintMode):integer;
begin
  case par of
   glNone:Result:=GL_DONT_CARE;
   glFast:Result:=GL_FASTEST;
   glbest:Result:=GL_NICEST;
   else Result:=GL_DONT_CARE;
  end;
end;

procedure TGLFrame.Process;
var i,par:integer;
begin
   glPushAttrib(GL_ALL_ATTRIB_BITS);
   glPushMatrix;
   {}
   glTranslatef(px,py,pz);
//   glScalef(fsx,fsy,fsz);
   {}
   trx:=trx+rx*incr;
   tryy:=tryy+ry*incr;
   trz:=trz+rz*incr;
   if trx>360 then trx:=trx-360;
   if tryy>360 then tryy:=tryy-360;
   if trz>360 then trz:=trz-360;
   if trx<0 then trx:=trx+360;
   if tryy<0 then tryy:=tryy+360;
   if trz<0 then trz:=trz+360;
   if trx<>0 then glRotateF(trx,1.0,0,0);
   if tryy<>0 then glRotateF(tryy,0,1.0,0);
   if trz<>0 then glRotateF(trz,0,0,1.0);
//   if Assigned(FFrames) then
   for i:=0 to FFrames.Count-1 do
     TGLFrame(FFrames[i]).Process;
   {}
    if Assigned(FTex) then
    begin
     glEnable(GL_TEXTURE_2D);
     if FAA then par:=GL_LINEAR else par:=GL_NEAREST;
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, par);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, par);
     glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                  Ftex.FW, FTex.FH,     //    
                  0, GL_RGBA, GL_UNSIGNED_BYTE, Ftex.Data);
     glHint(GL_PERSPECTIVE_CORRECTION_HINT,DParToGL(FPC));
     if FLS=glNone then glDisable(GL_LINE_SMOOTH) else glEnable(GL_POINT_SMOOTH);
     glHint(GL_LINE_SMOOTH_HINT,DParToGL(FLS));
     if FPC=glNone then glDisable(GL_POINT_SMOOTH) else glEnable(GL_POINT_SMOOTH);
     glHint(GL_POINT_SMOOTH_HINT,DParToGL(FPS));
     if FPolS=glNone then glDisable(GL_POLYGON_SMOOTH) else glEnable(GL_POINT_SMOOTH);
     glHint(GL_POLYGON_SMOOTH_HINT,DParToGL(FPolS));
    end;

  if FCompiled then glCallList(FList)
               else ProcessPolygons;
  glDisable(GL_TEXTURE_2D);
  glPopMatrix;
  glPopAttrib;
end;

constructor TGLTexture.Create;
begin
  fw:=-1;
  fh:=-1;
  MemUse:=-1;
  FrAss:=TList.Create;
  Ftr:=false;
  q:=True;
  FtrColor:=GLBlack;
end;

procedure TGLFrame.SetTexture(Texture:TGLTexture;DestroyOld:Boolean = true);
begin
  if DestroyOld then if Assigned(FTEx) then FTex.Destroy;
  if Assigned(Texture) then Texture.FrAss.Add(Self);
  FTex:=Texture;
end;

function MaxCr2(q:integer):integer;
var w:integer;
begin
  w:=2;
  while w<=q do w:=w*2;
  Result:=w div 2;
end;

procedure TGLTexture.Clear;
begin
  fw:=-1;
  fh:=-1;
  if memuse<>-1 then begin
     FreeMem(Data,MemUse);
     MemUse:=-1;
//     Data:=Nil;
  end;
//  SetLength(Data,0,0);
  FE:=True;
end;

procedure TGLTexture.LoadFromFile(FileName:string);
var q:TBitmap;
    i,j,incr,incr2:Integer;
    b:PByteArray;
    bb:byte;
    i1,i2:integer;
begin
  q:=TBitmap.Create;
  q.LoadFromFile(FileName);
  if q.Empty then begin q.Free;Exit; end;
//  b:=q.ScanLine[0];
//  incr:=High(b^) div q.Width;
  case q.PixelFormat of
   pf8bit:begin incr:=1;incr2:=0;end;
   pf24bit:begin incr:=3;incr2:=2;end;
   pf32bit:begin incr:=4;incr2:=2;end;
   else begin q.Free; Exit;end;
  end;
  FW:=MaxCr2(q.Width);
  FH:=MaxCr2(q.Height);
//  SetLength(Data,FW,FH);
  MemUse:=4*fw*fh;
//  Setlength(Data,MemUse);
  GetMem(Data,MemUse);
  i1:=0;i2:=0;
  for i:=0 to FH-1 do
   begin
     b:=q.ScanLine[i];
     for j:=0 to FW-1 do
     begin
       bb:=b^[j*incr+(incr2 div 3)];
       Data[i1,i2,2]:=bb;
       bb:=b^[j*incr+(incr2 div 2)];
       Data[i1,i2,1]:=bb;
       bb:=b^[j*incr+incr2];
       Data[i1,i2,0]:=bb;
       Data[i1,i2,3]:=255;
       inc(i2);if i2>(MaxTextureX-1) then begin i2:=0;inc(i1) end;
       if i1>(MaxTextureY-1) then begin i1:=0 end;
     end;
   end;
  FE:=False;
  q.Free;
end;

procedure TGLFrame.SetRotation(x,y,z,increment:GLFloat);
begin
  rx:=x;ry:=y;rz:=z;incr:=increment;
end;

function TGLViewPort.AddLight(Light:TGLLight):integer;
var q:PGLLight;
begin
  New(q);
  q^:=Light;
  Result:=FLights.Add(q);
end;

function TGLViewPort.AddLight(Light:PGLLight):integer;
begin
  Result:=FLights.Add(Light);
end;

procedure TGLFrame.SetPosition(x,y,z:GLFloat);
begin
  px:=x;
  py:=y;
  pz:=z;
end;

procedure TGLFrame.Move(dx,dy,dz:GLFloat);
begin
  px:=px+dx;
  py:=py+dy;
  pz:=pz+dz;
end;

destructor TGLViewPort.Destroy;
var i:integer;
begin
  for i:=0 to FLIghts.Count-1 do
  begin
    if Assigned(Flights.Items[i])
       then FreeMem(Flights.Items[i],SizeOf(TGLLight));
  end;
  FLights.Destroy;
  FMainFr.Destroy;
  inherited;
end;

destructor TGLFrame.Destroy;
var i:integer;
begin
  if Assigned(Ftex) then Ftex.Destroy;
  for i:=0 to FFrames.Count-1 do
   TGLFrame(FFrames[i]).Destroy;
  FFrames.Destroy; 
  inherited;
end;

procedure TGLViewPort.SetCameraPos(CamX,CamY, CamZ,TargetX,TargetY,TargetZ,UpX,UpY,UpZ:GLFloat);
begin
  FCamX:=CamX; FCamY:=CamY; FCamZ:=CamZ;
  FTGX:=TargetX; FTgY:=TargetY; FTgZ:=TargetZ;
  FUpX:=UpX; FUpY:=UpY; FUpZ:=UpZ;
end;

function Max(a,b:integer):integer;
begin
  if a>b then Result:=a else Result:=b;;
end;

procedure TGLFrame.AddFace(FVN:array of TFVN);
var i,n,m:integer;
    q:PGLFaceArr;
begin
  case High(FVN) of
   2:q:=@FFaces3V;
//   3:q:=@FFaces4V;
   else q:=@FFacesPol;
  end;
  n:=High(q^);
  inc(n);
  m:=0;
  for i:=0 to High(q^) do if High(q^[i])>m then m:=High(q^[i]);
  m:=max(High(FVN)+1,m);
  SetLength(q^,n+1,m);
  for i:=0 to High(FVN) do
   begin
    q^[n,i].Vertex:=FVN[i].Vertex;
    q^[n,i].Normal:=FVN[i].Normal;
    q^[n,i].TexX:=FVN[i].TexX;
    q^[n,i].TexY:=FVN[i].TexY;
   end;
end;

procedure TGLForm.Process;
begin
  if not FCanDraw then Exit;
  SendMessage(Handle,WM_PAINT,0,0);
end;

function GetL(i:integer):Cardinal;
begin
 result:=GL_LIGHT0+i;
end;

procedure TGLViewPort.Process;
var width,height,i:integer;
begin
  if FShMode=glSmooth then glShadeModel(GL_SMOOTH)
                      else glShadeModel(GL_FLAT);
  width:=FRect.Right-Frect.Left;
  Height:=Frect.Bottom-Frect.Top;
  glEnable(GL_SCISSOR_TEST);
  glScissor(FRect.Left,FRect.Top,FRect.Right,FRect.Bottom);
  glClearColor(FBKColor.r,FBKColor.g,FBKColor.b,FBKColor.a);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glDisable(GL_SCISSOR_TEST);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(FU, width/height, FNearZ, FFarZ);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  glViewport(FRect.Left,FRect.Top,FRect.Right,FRect.Bottom);
  gluLookAt(FCamX,FCamY,FCamZ,FTgX,FTgY,FTgZ,FUpX,FUpY,FUpZ);

   if Assigned(FLights) then
   for i:=0 to FLights.Count-1 do
    begin
      glEnable( GetL(i) );
      glLightModelF(GL_LIGHT_MODEL_AMBIENT,0.2);
      glLightModelF(GL_LIGHT_MODEL_LOCAL_VIEWER,1);
      glLightModelF(GL_LIGHT_MODEL_TWO_SIDE,0);
//      glLightfv( GetL(i), GL_LINEAR_ATTENUATION, @qq);
      glLightfv( GetL(i), GL_AMBIENT, @PGLLight(FLights[i]).Ambient.d[0]);
      glLightfv( GetL(i), GL_DIFFUSE, @PGLLight(FLights[i]).DIFFUSE.d[0] );
      glLightfv( GetL(i), GL_SPECULAR, @PGLLight(FLights[i]).Specular.d[0] );
      glLightfv( GetL(i), GL_POSITION, @PGLLight(FLights[i]).Position.x );
      glLightfv( GetL(i), GL_SPOT_DIRECTION, @PGLLight(FLights[i]).Direction.x );
    end;

  FMainFr.Process;
end;

constructor TGLFrame.Create(Parent:TGLFrame);
begin
  FParent:=Parent;
  FFrames:=TList.Create;
  if not Assigned(Parent) then Exit;
  Parent.FFrames.Add(self);
  FMat:=DefaultMaterial;
  FPC:=glBest;
  FPS:=glFast;
  FLS:=glFast;
  FPolS:=glFast;
  fsx:=1.0;
  fsy:=1.0;
  fsz:=1.0;
end;

procedure TGLForm.DeleteViewPort(ViewPort:TGLViewPort);
begin
  FViewPorts.Delete(FViewPorts.IndexOf(ViewPort));
end;

procedure TGLForm.DeleteViewPort(ViewPort:Integer);
begin
  FViewPorts.Delete(ViewPort);
end;

function TGLFrame.AddVertex(x,y,z:single):Integer;
var q:integer;
begin
  q:=High(FVertex);
  Inc(q);
  SetLength(FVertex,q+1);
  FVertex[q].x:=x;
  FVertex[q].y:=y;
  FVertex[q].z:=z;
  Result:=q;
end;

function TGLFrame.AddNormal(x,y,z:single):Integer;
var q:integer;
begin
  q:=High(FNormals);
  Inc(q);
  SetLength(FNormals,q+1);
  FNormals[q].x:=x;
  FNormals[q].y:=y;
  FNormals[q].z:=z;
  Result:=q;
end;

constructor TGLViewPort.Create(Form:TGLForm;Rect:TRect);
begin
  if not Assigned(Form) then begin raise Exception.Create('  '); Destroy; end;
  FLights:=TList.Create;
  FRect:=Rect;
  FBKColor:=GLBlack;
  FNearZ:=DefaultNearZ;
  FFarZ:=DefaultFarZ;
  FMainFr:=TGLFrame.Create(nil);
end;

function TGLForm.AddViewPort(Rect:TRect):integer;
begin
  Result:=FViewPorts.Add(TGLViewPort.Create(self,Rect));
end;

function TGLForm.GetViewPortCount:Integer;
begin
  Result:=FViewPorts.Count;
end;

function TGLForm.GetVP(index:integer):TGLViewPort;
begin
  Result:=FViewPorts[Index];
end;

procedure TGLForm.SetColorBits(Value:byte);
begin
  if Value in [8,16,24,32] then FColorBits:=Value else Raise Exception.Create('ColorBits must be 8, 16, 24 or 32!');
  if FASPF then SetDCPixelFormat;
end;

procedure TGLForm.SetDepthBits(Value:byte);
begin
  if Value in [8,16,24,32] then FDepthBits:=Value else Raise Exception.Create('DepthBits must be 8, 16, 24 or 32!');
  if FASPF then SetDCPixelFormat;
end;

constructor TGLForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if csDesigning in ComponentState then Exit;
  FColorBits:=16;
  FDepthBits:=16;
  DC := GetDC(Handle);
//  DC:=Canvas.Handle;
  SetDCPixelFormat;
  hrc := wglCreateContext(DC);
  wglMakeCurrent(DC, hrc);
  FViewPorts:=TList.Create;
//  glEnable( GL_CULL_FACE );
  glEnable( GL_ALPHA_TEST );
  glEnable( GL_LIGHTING );
  glEnable( GL_DEPTH_TEST );
  glEnable( GL_NORMALIZE );
  FCanDraw:=True;
end;

destructor TGLForm.Destroy;
var i:integer;
begin
 if csDesigning in ComponentState then begin inherited Destroy; Exit end;

 wglMakeCurrent(0, 0);
 wglDeleteContext(hrc);
 ReleaseDC(Handle, DC);
 if Assigned(FViewPorts) then
 for i:=0 to FViewPorts.Count-1 do
  TGLViewPort(FViewPorts[i]).Destroy;
 if Assigned(FViewPorts) then
 FViewPorts.Destroy;
 inherited Destroy;
end;

{=======================================================================
 WM_QUERYNEWPALETTE}
procedure TGLForm.WMQueryNewPalette(var Msg : TWMQueryNewPalette);
begin
 if (Palette <> 0) then begin
   Msg.Result := RealizePalette(DC);

 if (DWORD(Msg.Result) <> GDI_ERROR) then
   InvalidateRect(Handle, nil, False);
 end;
 Inherited;
end;

{=======================================================================
 WM_PALETTECHANGED}
procedure TGLForm.WMPaletteChanged(var Msg : TWMPaletteChanged);
begin
 if ((Palette <> 0) and (THandle(TMessage(Msg).wParam) <> Handle))
 then begin
   if (RealizePalette(DC) <> GDI_ERROR) then
     UpdateColors(DC);
   Msg.Result := 0;
 end;
 Inherited;
end;

procedure TGLForm.SetDCPixelFormat;
var
 hHeap: THandle;
 nColors, i: Integer;
 lpPalette: PLogPalette;
 byRedMask, byGreenMask, byBlueMask: Byte;
 nPixelFormat: Integer;
 pfd: TPixelFormatDescriptor;
begin
 FillChar(pfd, SizeOf(pfd), 0);

 with pfd do begin
   nSize     := sizeof(pfd);
   nVersion  := 1;
   dwFlags   := PFD_DRAW_TO_WINDOW or
                PFD_SUPPORT_OPENGL or
                PFD_DOUBLEBUFFER;
   iPixelType:= PFD_TYPE_RGBA;
   cColorBits:= FColorBits;
{   cRedBits:=8;
   cRedShift:=0;
   cGreenBits:=8;
   cGreenShift:=0;
   cBlueBits:=8;
   cBlueShift:=0; }
   cAccumBits:=16;
   cAccumRedBits:=5;
   cAccumGreenBits:=5;
   cAccumBlueBits:=5;
   cAccumAlphaBits:=1;
   cStencilBits:=16;
   cDepthBits:= 16{FDepthBits};
   iLayerType:= PFD_MAIN_PLANE;
 end;
 nPixelFormat := ChoosePixelFormat(DC, @pfd);
 SetPixelFormat(DC, nPixelFormat, @pfd);

 DescribePixelFormat(DC, nPixelFormat, sizeof(TPixelFormatDescriptor), pfd);
 if ((pfd.dwFlags and PFD_GENERIC_FORMAT) = 0) then
 if ((pfd.dwFlags and PFD_GENERIC_ACCELERATED) = 0) then
     ShowMessage('!!!');

 if ((pfd.dwFlags and PFD_NEED_PALETTE) <> 0) then begin
   nColors   := 1 shl pfd.cColorBits;
   hHeap     := GetProcessHeap;
   lpPalette := HeapAlloc(hHeap, 0, sizeof(TLogPalette) + (nColors * sizeof(TPaletteEntry)));

   lpPalette^.palVersion := $300;
   lpPalette^.palNumEntries := nColors;

   byRedMask   := (1 shl pfd.cRedBits) - 1;
   byGreenMask := (1 shl pfd.cGreenBits) - 1;
   byBlueMask  := (1 shl pfd.cBlueBits) - 1;

   for i := 0 to nColors - 1 do begin
     lpPalette^.palPalEntry[i].peRed   := (((i shr pfd.cRedShift)   and byRedMask)   * 255) DIV byRedMask;
     lpPalette^.palPalEntry[i].peGreen := (((i shr pfd.cGreenShift) and byGreenMask) * 255) DIV byGreenMask;
     lpPalette^.palPalEntry[i].peBlue  := (((i shr pfd.cBlueShift)  and byBlueMask)  * 255) DIV byBlueMask;
     lpPalette^.palPalEntry[i].peFlags := 0;
   end;

   Palette := CreatePalette(lpPalette^);
   HeapFree(hHeap, 0, lpPalette);

   if (Palette <> 0) then begin
     SelectPalette(DC, Palette, False);
     RealizePalette(DC);
   end;
 end;
end;

procedure TGLForm.WMPaint(var Msg:TMessage);
var i:integer;
begin
  inherited;
  FCanDraw:=False;
  for i:=0 to FViewPorts.Count-1 do
    begin
      TGLViewPort(FViewPorts[i]).Process;
    end;
//  glFlush;
  glFinish;
  SwapBuffers(DC);
  FCanDraw:=True;
end;

function TGLFrame.GetFrCount:integer;
begin
  Result:=FFrames.Count;
end;

function TGLFrame.GetFr(index:integer):TGLFrame;
begin
  Result:=FFrames.Items[index];
end;

function TGLViewPort.GetLgCount:integer;
begin
  Result:=FLights.Count;
end;

function TGLViewPort.GetLg(index:integer):PGLLight;
begin
  Result:=FLights.Items[index];
end;

function TGLForm.NewViewPort(Rect:TRect):TGLViewPort;
var q:TGLViewPort;
begin
  q:=TGLViewPort.Create(self,Rect);
  FViewPorts.Add(q);
  Result:=q;
end;

function TGLFrame.AddFrame(Frame:TGLFrame):integer;
begin
  if not Assigned(Frame) then begin Result:=1;Exit; end;
  Result:=FFrames.Add(Frame);
//  Frame.FParent:=Self;
end;

procedure TGLFrame.Rotate(x,y,z:GLFloat);
begin
  rx:=x;ry:=y;rz:=z;incr:=0;
end;

destructor TGLTexture.Destroy;
var i:integer; 
begin
  if MemUse<>-1 then Clear;
  for i:=0 to FrAss.Count-1 do
   if Assigned(FrAss.Items[i]) then TGLFrame(FrAss.Items[i]).FTex:=nil;
  FrAss.Destroy;
  inherited;
end;

{procedure TGLFrame.Scale(sx,sy,sz:GLFloat);
begin
  fsx:=sx;
  fsy:=sy;
  fsz:=sz;
end;}

end.
